"""
pyplayer.py is a Player client written in/for python
"""

import socket
import thread
import time
from ppmessage import *


class player:
    """
    wrapper class for communication to a player server.
    """

    # sensory data repository
    misc	= {}	# misc
    gripper	= {}	# gripper
    position	= {}	# position
    sonar	= {}	# sonar
    laser	= {}	# laser
    vision	= {}	# vision
    ptz		= {}	# ptz
    audio	= {}	# audio
    laserbeacon	= {}	# laserbeacon
    broadcast	= {}	# broadcast
    speech	= {}	# speech
    gps		= {}	# gps
    bps         = {}	# bps


    def __init__(self, hostname='localhost', port=6665, auto=1, speed=20): #---
	"""
	connect to the specified player server, and
	request a player device.
	"""
	try:
	    # make a connection to the specified player server
	    self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	    self.__socket.connect((hostname, port))
	    pversion = self.__nrecv(32)		# version string
	    self.__socket.setblocking(0)
	    # create automatic update thread
	    self.lock = thread.allocate_lock()
	    if auto: self.start_update(speed)
	    else: self.stop_update()
	except socket.error, errstr:
	    self.__socket = None
	    raise 'cannot create a socket: %s' % errstr
	except thread.error:
	    raise 'cannot run a update thread.'
	else:
	    # save the information of the connection
	    self.server = (hostname, port)
	    print '<connected to %s:%d>' % self.server + ' (%s)' % pversion


    def close(self): #---------------------------------------------------------
	"""
	disconnect from the player server.
	"""
	# close the connection
	if self.__socket is not None:
	    try:
	        self.__update_flag = 0
	        self.__socket.shutdown(2)
		self.__socket.close()
	    except socket.error, errstr:
	        raise 'cannot close the socket: %s' % errstr
	    else:
		print '<connection to %s:%d is closed>' % self.server


    def __repr__(self): #------------------------------------------------------
        """
	convert itself to a string.
	"""
	return 'player client connected to %s:%d' % self.server


    def __wait_recv(self, size): #---------------------------------------------
        """
	receive the specified number of bytes from the server.
	if data is not available yet, wait forever.
	"""
	recv_str = ''
	while len(recv_str) < size:
	    try:
		recv_str += self.__socket.recv(size - len(recv_str))
	    except socket.error:
	        time.sleep(0.01)	# wait more data
	return recv_str


    def __wait_packet(self): #-------------------------------------------------
        """
	receive a packet, which is a minimal, meaningful data chunk.
	if a packet is not available yet, wait forever.
	"""
	# receive header
	header = self.__wait_recv(32)
	hdrinfo, size = unpack_header(header)
	# receive payload
	payload = self.__wait_recv(size)
	# return only useful data
	return (hdrinfo, payload)


    def __wait_response(self, device): #---------------------------------------
        """
	wait a response from the player server.
	ignore all irrelevant packets until it gets response from the server.
	"""
	# wait the right response
	while 1:
	    header, payload = self.__wait_packet()
	    if header[0] == response and header[1] == devicenum[device[0]] and header[2] == device[1]:
	        break
	# return the payload
	return payload


    def __nsend(self, data): #-------------------------------------------------
        """
	send the data string to the server.
	caller should catch exceptions for proper error handling,
	i.e. broken pipe.
	"""
	sent = 0
	while sent < len(data):
	    sent += self.__socket.send(data[sent:])


    def __nrecv(self, size): #-------------------------------------------------
        """
	receive the specified number of bytes from the server.
	caller should catch exceptions for proper error handling
	"""
	recv_str = ''
	while len(recv_str) < size:
	    try:
		recv_str += self.__socket.recv(size - len(recv_str))
	    except socket.error, (errno, errstr):
	        if errno == 11 and recv_str == '':
		    return None	# network link is fine, but data is not ready
		else:
		    raise	# network link got broken
	return recv_str


    def __recv_packet(self): #-------------------------------------------------
        """
	receive a packet, which is a minimal, meaningful data chunk.
	caller should catch exceptions for proper error handling
	"""
	# receive header
	header = self.__nrecv(32)
	if header == None: return None		# no data is available yet
	hdrinfo, size = unpack_header(header)
	# receive payload
	payload = self.__wait_recv(size)
	# return only useful data
	return (hdrinfo, payload)


    def start(self, device, index=0, mode='r'): #----------------------------
        """
	send request to read device.
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'player', size=7)
	    payload = pack_player_request(device, index, mode)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    payload = self.__wait_response(('player', 0))
	    if unpack_player_request(payload) == 'e':
		raise 'device request error: (%s, %d)' % (device, index)
	    else:
	        print '<device %s:%d is enabled>' % (device, index)
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise
	else:
	    # re-start the update thread
	    self.lock.release()


    def stop(self, device, index=0): #--------------------------------------
        """
	send request to close device.
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'player', size=7)
	    payload = pack_player_request(device, index, 'c')
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    payload = self.__wait_response(('player', 0))
	    if unpack_player_request(payload) == 'e':
	        raise 'error'
	    else:
	        print '<device %s:%d is disabled>' % (device, index)
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'device close error: (%s, %d)' % (device, index)
	else:
	    # re-start the update thread
	    self.lock.release()


    def fetch_data(self): #----------------------------------------------------
        """
	fetch one round of data.
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'player', size=2)
	    payload = pack_player_fetch()
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('player', 0))
	    print '<received one round of data>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot fetch data'
	else:
	    # re-start the update thread
	    self.lock.release()


    def set_mode(self, mode): #---------------------------------------------
        """
	change data mode.

	when the mode is zero, the player server feeds data continuously.
	when the mode is non-zero, the player server sends one round of data
	only when there is a request from users.
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'player', size=3)
	    payload = pack_player_mode(mode)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('player', 0))
	    if mode==0: print '<player is in continuous mode>'
	    else: print '<player is in request/reply mode>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot change data mode (%d)' % mode
	else:
	    # re-start the update thread
	    self.lock.release()


    def set_frequency(self, frequency): #---------------------------------
        """
	change data feeding frequency.
	
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'player', size=4)
	    payload = pack_player_frequency(frequency)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('player', 0))
	    print '<data frequency is %dHz now>' % frequency
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot set data frequency (%d)' % frequency
	else:
	    # re-start the update thread
	    self.lock.release()


    def __gripper_command(self, gripcommand, press=1, carry=1, gripindex=0):
	"""
	send a gripper command to the player server.
	"""
	try:
	    # send the command
	    header = pack_header(command, 'gripper', gripindex, size=2)
	    if command == gripper_command['press']:
		payload = pack_gripper_command(gripcommand, press)
	    elif command == gripper_command['carry']:
		payload = pack_gripper_command(gripcommand, carry)
	    else:
		payload = pack_gripper_command(gripcommand, 0)
	    self.__nsend(header+payload)
	    # NO RESPONSE
	    print '<sent the command (%s)>' % gripcommand
	except:
	    raise 'cannot send the command (%s)' % gripcommand


    def gripper_open(self, index=0):
        """
	send the 'open' gripper command to the player server.
	"""
	self.__gripper_command('open', gripindex=index)


    def gripper_close(self, index=0):
        """
	send the 'close' gripper command to the player server.
	"""
	self.__gripper_command('close', gripindex=index)


    def gripper_stop(self, index=0):
        """
	send the 'stop' gripper command to the player server.
	"""
	self.__gripper_command('stop', gripindex=index)


    def gripper_up(self, index=0):
        """
	send the 'up' gripper command to the player server.
	"""
	self.__gripper_command('up', gripindex=index)


    def gripper_down(self, index=0):
        """
	send the 'down' gripper command to the player server.
	"""
	self.__gripper_command('down', gripindex=index)


    def gripper_stay(self, index=0):
        """
	send the 'stay' gripper command to the player server.
	"""
	self.__gripper_command('stay', gripindex=index)


    def gripper_store(self, index=0):
        """
	send the 'store' gripper command to the player server.
	"""
	self.__gripper_command('store', gripindex=index)


    def gripper_deploy(self, index=0):
        """
	send the 'deploy' gripper command to the player server.
	"""
	self.__gripper_command('deploy', gripindex=index)


    def gripper_halt(self, index=0):
        """
	send the 'halt' gripper command to the player server.
	"""
	self.__gripper_command('halt', gripindex=index)


    def gripper_press(self, press_time=1, index=0):
        """
	send the 'press' gripper command to the player server.
	"""
	self.__gripper_command('press', press=press_time, gripindex=index)


    def gripper_carry(self, carry_time=1, index=0):
        """
	send the 'carry' gripper command to the player server.
	"""
	self.__gripper_command('carry', carry=carry_time, gripindex=index)


    def set_speed(self, speed=None, turnrate=None, sidespeed=None, index=0):
	"""
	change the speed and turnrate of wheels.
	If a new speed or a new turnate is None, the current value will be used.
	"""
	try:
	    # set proper speed and turnrate
	    if speed is None: speed = self.position[index][3]
	    if sidespeed is None: sidespeed = self.position[index][4]
	    if turnrate is None: turnrate = self.position[index][5]
	    # send command
	    header = pack_header(command, 'position', index, size=6)
	    payload = pack_position_speed(speed, sidespeed, turnrate)
	    self.__nsend(header+payload)
	    # NO RESPONSE
	    print '<set the speed to (%d,%d,%d)>' % (speed, sidespeed, turnrate)
	except:
	    raise 'cannot set position speed (%d,%d,%d)' % (speed, sidespeed, turnrate)


    def turnon_motor(self): #-------------------------------------------------
        """
	turon on motors.
	
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'position', size=2)
	    payload = pack_position_motors(1)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('position', 0))
	    print '<turned on the motors>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot turn on the motors'
	else:
	    # re-start the update thread
	    self.lock.release()


    def turnoff_motor(self): #-------------------------------------------------
        """
	turon motors off.
	
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'position', size=2)
	    payload = pack_position_motors(0)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('position', 0))
	    print '<turned off motors>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot turn off the motors'
	else:
	    # re-start the update thread
	    self.lock.release()


    def set_control(self, mode='separate'): #----------------------------------
        """
	set velocity control mode.
	
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'position', size=2)
	    if mode == 'separate': vc = 1
	    else: vc = 0
	    payload = pack_position_velocity(vc)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('position', 0))
	    print '<set the velocity control to ' + mode + '>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot set the velocity control'
	else:
	    # re-start the update thread
	    self.lock.release()


    def reset_odometry(self): #------------------------------------------------
	"""
	reset odometry information to (0,0,0).
	
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'position', size=1)
	    payload = pack_position_reset()
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('position', 0))
	    print '<reset the odometry>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot reset the odometry'
	else:
	    # re-start the update thread
	    self.lock.release()


    def turnon_sonar(self): #-------------------------------------------------
        """
	turon on sonar transducers.
	
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'sonar', size=2)
	    payload = pack_sonar_state(1)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('sonar', 0))
	    print '<turned on the sonar transducers>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot turn on the sonar transducers'
	else:
	    # re-start the update thread
	    self.lock.release()


    def turnoff_sonar(self): #-------------------------------------------------
        """
	turon off sonar transducers.
	
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'sonar', size=2)
	    payload = pack_sonar_state(0)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('sonar', 0))
	    print '<turned off the sonar transducers>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot turn off the sonar transducers'
	else:
	    # re-start the update thread
	    self.lock.release()


    def configure_laser(self, min=-90.0, max=90.0, resolution=0.5, intensity=1): #---
        """
	change the configuration of laser sensors.
	
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'laser', size=7)
	    payload = pack_laser(min, max, resolution, intensity)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('laser', 0))
	    print '<changed the laser configuration>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot change the laser configuration'
	else:
	    # re-start the update thread
	    self.lock.release()


    def beep(self, frequency=1000, amplitude=300, duration=1): #---------------
        """
	play a beep sound.
	"""
	try:
	    # send command
	    header = pack_header(command, 'audio', size=6)
	    payload = pack_audio(frequency, amplitude, duration)
	    self.__nsend(header+payload)
	    # NO RESPONSE
	    print '<sent a beep command>'
	except:
	    raise 'cannot send a beep command'


    def set_laserbeacon_bit(self, bit_count=8, bit_width=50): #----------------
        """
	configure laserbeacon device (bit).
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'laserbeacon', size=4)
	    payload = pack_laserbeacon_bit(bit_count, bit_width)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('laserbeacon', 0))
	    print '<change the laserbeacon configuration (bit)>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot change the laserbeacon configuration (bit)'
	else:
	    # re-start the update thread
	    self.lock.release()


    def set_laserbeacon_threshold(self, zero_threshold=60, one_threshold=60): #----------------
        """
	configure laserbeacon device (threshold).
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'laserbeacon', size=5)
	    payload = pack_laserbeacon_threshold(zero_threshold, one_threshold)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('laserbeacon', 0))
	    print '<change the laserbeacon configuration (threshold)>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot change the laserbeacon configuration (threshold)'
	else:
	    # re-start the update thread
	    self.lock.release()


    def set_ptz(self, pan, tilt, zoom): #--------------------------------------------
        """
	move a camera.
	"""
	try:
	    # send command
	    header = pack_header(command, 'ptz', size=6)
	    payload = pack_ptz(pan, tilt, zoom)
	    self.__nsend(header+payload)
	    # NO RESPONSE
	    print '<sent a ptz command>'
	except:
	    raise 'cannot send a ptz command'


    def send_message(self, message): #--------------------------------------------
        """
	broadcast the message over the wireless network.
	"""
	try:
	    # send command
	    header = pack_header(command, 'broadcast', size=2+len(message))
	    payload = pack_broadcast(message)
	    self.__nsend(header+payload)
	    # NO RESPONSE
	    print '<sent a broadcast command>'
	except:
	    raise 'cannot send a broadcast command'


    def speak(self, message): #------------------------------------------------
        """
	speak the message.
	"""
	try:
	    # send command
	    header = pack_header(command, 'speech', size=256)
	    payload = pack_speech(message)
	    self.__nsend(header+payload)
	    # NO RESPONSE
	    print '<sent a speech command>'
	except:
	    raise 'cannot send a speech command'


    def teleport(self, x=0, y=0, heading=0): #----------------------------------
        """
	teleport the robot to a new location.
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'gps', size=12)
	    payload = pack_audio(x, y, heading)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('gps', 0))
	    print '<sent a gps request>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot send a gps request'
	else:
	    # re-start the update thread
	    self.lock.release()


    def set_bps_gain(self, gain): #--------------------------------------------
        """
	change the gain value of BPS devices.
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'bps', size=5)
	    payload = pack_bps_gain(gain)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('bps', 0))
	    print '<sent a bps gain request>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot send a bps gain request'
	else:
	    # re-start the update thread
	    self.lock.release()


    def add_laserbeacon(self, id, px, py, pa, ux, uy, ua): #-------------------
        """
	add a piece of laserbeacon information.
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'bps', size=26)
	    payload = pack_bps_beacon(id, px, py, pa, ux, uy, ua)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('bps', 0))
	    print '<sent a bps laserbeacon request>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot send a bps laserbeacon request'
	else:
	    # re-start the update thread
	    self.lock.release()


    def set_laser_position(self, px, py, pa): #--------------------------------
        """
	set the position of a laser sensor in a robot body.
	"""
	try:
	    # stop the update thread
	    self.lock.acquire()
	    # send request
	    header = pack_header(request, 'bps', size=13)
	    payload = pack_bps_laser(px, py, pa)
	    self.__nsend(header+payload)
	    # wait a respose from the player server
	    self.__wait_response(('bps', 0))
	    print '<sent a bps laser position request>'
	except:
	    # re-start the update thread
	    self.lock.release()
	    raise 'cannot send a bps laser position request'
	else:
	    # re-start the update thread
	    self.lock.release()


    def __update_thread(self): #-----------------------------------------------
        """
	update sensory information.
	"""
	print '<started a thread updating sensory information>'
	while self.__update_flag:
	    self.update()
	    time.sleep(1.0 / self.__update_speed)
	print '<stop the update thread>'


    def start_update(self, speed=20): #----------------------------------------
        """
	start to update sensory information continuously.
	"""
	self.__update_flag = 1
	self.__update_speed = speed
	thread.start_new_thread(self.__update_thread, ())


    def stop_update(self): #---------------------------------------------------
        """
	stop updating sensory information continuously.
	"""
	self.__update_flag = 0


    def update(self): #--------------------------------------------------------
        """
	update sensory information.
	"""
	try:
	    # for synchronization with other functions
	    self.lock.acquire()
	    # update information
	    while 1:
	        # read a packet
	        packet = self.__recv_packet()
		if packet == None: break
		(type, device, index), payload = packet
		# update information
		if device == 0x0001:	# player
		    print '<unexpected packet (%s:%d)>' % (devicestr[device], index)
		elif device == 0x0002:	# misc
		    self.misc[index] = unpack_misc(payload)
		elif device == 0x0003:	# gripper
		    self.gripper[index] = unpack_gripper(payload)
		elif device == 0x0004:	# position
		    self.position[index] = unpack_position(payload)
		elif device == 0x0005:	# sonar
		    self.sonar[index] = unpack_sonar(payload)
		elif device == 0x0006:	# laser
		    self.laser[index] = unpack_laser(payload)
		elif device == 0x0007:	# vision
		    self.vision[index] = unpack_vision(payload)
		elif device == 0x0008:	# ptz
		    self.ptz[index] = unpack_ptz(payload)
		elif device == 0x0009:	# audio
		    self.audio[index] = unpack_audio(payload)
		elif device == 0x000A:	# laserbeacon
		    self.laserbeacon[index] = unpack_laserbeacon(payload)
		elif device == 0x000B:	# broadcast
		    self.broadcast[index] = unpack_broadcast(payload)
		elif device == 0x000C:	# speech
		    print '<speech device is not supposed send data back>'
		elif device == 0x000D:	# gps
		    self.gps[index] = unpack_gps(payload)
		elif device == 0x0010:	# bps
		    self.bps[index] = unpack_bps(payload)
		else:			# unknown device
		    print '<unknown device (%d:%d) packet>' % (device, index)
	except socket.error:
	    # for synchronization with other functions
	    self.lock.release()
	    raise 'network connection error: (%s, %d)' % (device, index)
	else:
	    # for synchronization with other functions
	    self.lock.release()


    def __byte_to_reversed_list(self, byte): #------------------------------------------
        """
	convert a byte into a bit list.
	"""
	result = []
	for idx in range(7,-1,-1):
	    result.append(byte & 0x01)
	    byte >>= 1
	return result


    def get_misc(self, index=0): #---------------------------------------------
        """
	return the sensory value of a misc device.
	"""
	fbs = self.__byte_to_reversed_list(self.misc[index][0])[:5]
	fbs.reverse()
	rbs = self.__byte_to_reversed_list(self.misc[index][1])[:5]
	rbs.reverse()
	digin = self.__byte_to_reversed_list(self.misc[index][4])
	digin.reverse()
	return ( tuple(fbs), tuple(rbs),
		 self.misc[index][2] / 10.0,
		 self.misc[index][3],
		 tuple(digin) )


    def get_gripper(self, index=0): #------------------------------------------
        """
	return the sensory value of a gripper device.
	"""
	state = self.__byte_to_reversed_list(self.gripper[index][0])
	beams = self.__byte_to_reversed_list(self.gripper[index][1])[:6]
	return ( tuple(state), tuple(beams) )


    def get_position(self, index=0): #-----------------------------------------
        """
	return the sensory value of a position device.
	"""
	return self.position[index]


    def get_sonar(self, index=0): #--------------------------------------------
        """
	return the sensory value of a sonar device.
	"""
	return self.sonar[index]


    def get_laser(self, index=0): #--------------------------------------------
        """
	return the sensory value of a laser device.
	"""
	return ( ( self.laser[index][0]/100.0, self.laser[index][1]/100.0,
		   self.laser[index][2]/100.0, self.laser[index][3] ),
		 tuple( self.laser[index][4:(self.laser[index][3]+4)] ) )


    def get_vision(self, index=0): #-------------------------------------------
        """
	return the sensory value of a vision device.
	"""
	return self.vision[index]


    def get_ptz(self, index=0): #----------------------------------------------
        """
	return the sensory value of a ptz device.
	"""
	return self.ptz[index]


    def get_audio(self, index=0): #--------------------------------------------
        """
	return the sensory value of a audio device.
	"""
	return ( tuple(self.audio[index][0:2]),
	         tuple(self.audio[index][2:4]),
	         tuple(self.audio[index][4:6]),
	         tuple(self.audio[index][6:8]),
	         tuple(self.audio[index][8:10]) )


    def get_laserbeacon(self, index=0): #--------------------------------------
        """
	return the sensory value of a laserbeacon device.
	"""
	lb = []
	for idx in range(1, self.laserbeacon[index][0]*4+1, 4):
	    lb.append(tuple(self.laserbeacon[index][idx:idx+4]))
	return tuple(lb)


    def get_broadcast(self, index=0): #----------------------------------------
        """
	return the sensory value of a broadcast device.
	"""
	return self.broadcast[index][0]


    def get_gps(self, index=0): #----------------------------------------------
        """
	return the sensory value of a gps device.
	"""
	return self.gps[index]


    def get_bps(self, index=0): #----------------------------------------------
        """
	return the sensory value of a bps device.
	"""
	return ( tuple(self.bps[index][0:3]), tuple(self.bps[index][3:6]), self.bps[index][6] )


    def dump_sensors(self): #--------------------------------------------------
        """
	print sensory information.
	"""
	print self.misc
	print self.gripper	
	print self.position
	print self.sonar
	print self.laser
	print self.vision
	print self.ptz
	print self.audio
	print self.laserbeacon
	print self.broadcast
	print self.gps
	print self.bps

